import { type ByteRange } from './byte-range.ts';
/**
 * A streaming interface for blob/file content.
 */
export interface LazyContent {
    /**
     * The total length of the content.
     */
    byteLength: number;
    /**
     * Returns a stream that can be used to read the content. When given, the `start` index is
     * inclusive indicating the index of the first byte to read. The `end` index is exclusive
     * indicating the index of the first byte not to read.
     */
    stream(start?: number, end?: number): ReadableStream<Uint8Array>;
}
export interface LazyBlobOptions {
    /**
     * The range of bytes to include from the content. If not specified, all content is included.
     */
    range?: ByteRange;
    /**
     * The MIME type of the content. The default is an empty string.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/Blob#type)
     */
    type?: string;
}
/**
 * A [`Blob`](https://developer.mozilla.org/en-US/docs/Web/API/Blob) that may be backed by a stream
 * of data. This is useful for working with large blobs that would be impractical to load into
 * memory all at once.
 *
 * This class is an extension of JavaScript's built-in `Blob` class with the following additions:
 *
 * - The constructor may accept a `LazyContent` object instead of a `BlobPart[]` array
 * - The constructor may accept a `range` in the options to specify a subset of the content
 *
 * In normal usage you shouldn't have to specify the `range` yourself. The `slice()` method
 * automatically takes care of creating new `LazyBlob` instances with the correct range.
 *
 * [MDN `Blob` Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob)
 */
export declare class LazyBlob extends Blob {
    #private;
    constructor(parts: BlobPart[] | LazyContent, options?: LazyBlobOptions);
    /**
     * Returns the blob's contents as an [`ArrayBuffer`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/ArrayBuffer).
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/arrayBuffer)
     */
    arrayBuffer(): Promise<ArrayBuffer>;
    /**
     * Returns the blob's contents as a byte array.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/bytes)
     */
    bytes(): Promise<Uint8Array>;
    /**
     * The size of the blob in bytes.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/size)
     */
    get size(): number;
    /**
     * Returns a new [`Blob`](https://developer.mozilla.org/en-US/docs/Web/API/Blob) that contains the data in the specified range.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/slice)
     */
    slice(start?: number, end?: number, contentType?: string): Blob;
    /**
     * Returns a stream that can be used to read the blob's contents.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/stream)
     */
    stream(): ReadableStream<Uint8Array>;
    /**
     * Returns the blob's contents as a string.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/text)
     */
    text(): Promise<string>;
}
export interface LazyFileOptions extends LazyBlobOptions {
    /**
     * The last modified timestamp of the file in milliseconds. Defaults to `Date.now()`.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/File/File#lastmodified)
     */
    lastModified?: number;
}
/**
 * A [`File`](https://developer.mozilla.org/en-US/docs/Web/API/File) that may be backed by a stream
 * of data. This is useful for working with large files that would be impractical to load into
 * memory all at once.
 *
 * This class is an extension of JavaScript's built-in `File` class with the following additions:
 *
 * - The constructor may accept a `LazyContent` object instead of a `BlobPart[]` array
 * - The constructor may accept a `range` in the options to specify a subset of the content
 *
 * In normal usage you shouldn't have to specify the `range` yourself. The `slice()` method
 * automatically takes care of creating new `LazyBlob` instances with the correct range.
 *
 * [MDN `File` Reference](https://developer.mozilla.org/en-US/docs/Web/API/File)
 */
export declare class LazyFile extends File {
    #private;
    constructor(parts: BlobPart[] | LazyContent, name: string, options?: LazyFileOptions);
    /**
     * Returns the file's content as an [`ArrayBuffer`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/ArrayBuffer).
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/arrayBuffer)
     */
    arrayBuffer(): Promise<ArrayBuffer>;
    /**
     * Returns the file's contents as a byte array.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/bytes)
     */
    bytes(): Promise<Uint8Array>;
    /**
     * The size of the file in bytes.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/size)
     */
    get size(): number;
    /**
     * Returns a new [`Blob`](https://developer.mozilla.org/en-US/docs/Web/API/Blob) that contains the data in the specified range.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/slice)
     */
    slice(start?: number, end?: number, contentType?: string): Blob;
    /**
     * Returns a stream that can be used to read the file's contents.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/stream)
     */
    stream(): ReadableStream<Uint8Array>;
    /**
     * Returns the file's contents as a string.
     *
     * [MDN Reference](https://developer.mozilla.org/en-US/docs/Web/API/Blob/text)
     */
    text(): Promise<string>;
}
//# sourceMappingURL=lazy-file.d.ts.map